#!/bin/sh
# The next line is executed by /bin/sh, but not tcl \
exec wish "$0" -- $*

set TauVersion	"Tau 0.4 - 12 February 2015"

#  Tau -- a gui for tiny automata verification
#  (c) 2015 Gerard J. Holzmann
#  California Institute of Technology, Pasadena, CA, USA.
#
#  This is a teaching tool created for use in Caltech Course
#  CS118: Logic Model Checking for Formal Software Verification 
#
#  Tau is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# TODO:
# remnants of old process names or states can still appear
#   in saved files (after renaming and further editing)
# states are referened by nr, S$nr, and by name
#   should consistently be nr only

# colornames:
# palegreen, wheat, snow, khaki1, rosybrown1, plum1,
# gray97, thistle1, lavenderblush

set CW	1000
set CH	 600
set Fname ""

set SW	250	;# default horizontal separation between states
set SH	106	;# default vertical seperation between states
set WW	 40	;# state text window width

set ShW  50	;# state circle width
set ShH  50	;# state circle height

set SCX 60	;# default x-location for state creation
set SCY 60	;# default y-location for state creation

set cnt 0	;# for internal numbering of states
set curproc "p0"
set which_state 0
set which_x 0
set which_y 0
set next_state 0
set with_shadows 1
set shadow_col grey80
set state_shape oval
set shadow_dist 4
set with_scroll_bars 0	;# on left-side simulation output window
set nowarn 0
set anychange 0

set Helv	-Adobe-Helvetica-Medium-R-Normal--*-110-*-*-*-*-*-*
set Bcol	honeydew	;# bg color buttons

wm title . "Tau - a Tool for Tiny Automata"
wm geometry . ${CW}x${CH}+50+50

frame .b -bg honeydew	;# buttons
pack .b -side left -fill y ;# -expand yes

set BW 10

	frame .b.top -bg ivory
	label .b.top.fl -relief raised -bg ivory \
		-width $BW -text "File Menu"
	button .b.top.help -relief raised -bg ivory \
		-width $BW -text "Help" -command { do_help }
	bind .b.top.fl <ButtonPress-1> \
		"tk_popup .b1 %X %Y"
	  pack append .b.top \
		.b.top.fl { left fillx filly expand } \
		.b.top.help { left fillx filly expand }

	frame .b.c
	frame .b.c.r1
	  button .b.c.r1.ns -text "+state" \
		-bg $Bcol \
		-width $BW -command { incr cnt; new_state $cnt }
	  button .b.c.r1.va -text "+context"  \
		-bg $Bcol \
		-width $BW -command { add_data }

	  pack append .b.c.r1 \
		.b.c.r1.ns {left fillx expand} \
		.b.c.r1.va {right fillx expand}

	frame .b.c.r2
	  button .b.c.r2.sm -text "+sim"   \
		-bg $Bcol \
		-width $BW -command { do_sim }
	  button .b.c.r2.rn -text "+verify"\
		-bg $Bcol \
		-width $BW -command { do_check }
	  pack append .b.c.r2 \
		.b.c.r2.sm {left} \
		.b.c.r2.rn {right}

	frame .b.c.r3
	  button .b.c.r3.gm -text "+model" \
		-bg $Bcol \
		-width $BW -command { do_pml 2 }
	  button .b.c.r3.rs -text "+reset" \
		-bg $Bcol \
		-width $BW -command { reset_states }
	  pack append .b.c.r3 \
		.b.c.r3.gm {left} \
		.b.c.r3.rs {right}

	pack append .b.c \
		.b.c.r1 {top fillx expand} \
		.b.c.r2 {top fillx expand} \
		.b.c.r3 {top fillx expand}

	frame .b.notes -width 10 -bg ivory
	frame .b.log -width 10 -bg ivory

	pack append .b \
		.b.top   { top fillx filly } \
		.b.c     { top fillx } \
		.b.notes { top fillx filly expand} \
		.b.log { top fillx filly }

	text .b.log.t -width 30 -height 5 -bg honeydew
	pack append .b.log \
		.b.log.t { top fillx filly expand }
	if {$with_scroll_bars} {
		text .b.notes.t -width 30 -yscrollcommand {.b.notes.v set}
		scrollbar .b.notes.v \
			-bg honeydew \
			-orient vert \
			-command {.b.notes.t yview }
		pack append .b.notes \
			.b.notes.t {right fillx filly expand} \
			.b.notes.v {right filly expand}
	} else {
		text .b.notes.t -width 20
		pack append .b.notes \
			.b.notes.t { right fillx filly expand }
	}

frame .e -bg white	;# edit area
pack .e -side top -fill both -expand yes
	set cv [canvas .e.c  -relief raised -bd 2\
		-scrollregion {-1c -1c 100c 100c} \
		-background ivory \
		-xscrollcommand ".e.hscroll set" \
		-yscrollcommand ".e.vscroll set"]
	scrollbar .e.vscroll -relief sunken \
		-command ".e.c yview"
	scrollbar .e.hscroll -relief sunken -orient horiz \
		-command ".e.c xview"
	pack append .e \
		.e.vscroll {right filly} \
		.e.hscroll {bottom fillx} \
		.e.c {top expand fill}

	menu .b1 -tearoff 0 -bg ivory -fg black
	.b1 add command -label "New"     -command { start_over }
	.b1 add command -label "Open"    -command { do_open }
	.b1 add command -label "Save"    -command { do_save 0 }
	.b1 add command -label "Save As" -command { do_save 1 }
.b1 add command -label "Align" -command { align_grid }
	.b1 add command -label "Exit"    -command { do_exit   }

	menu .b2 -tearoff 0
	.b2 add command -label "Set as initial state; name process" \
		-command { set_initial $which_state }
	.b2 add command -label "Duplicate all States Reachable from this State" \
		-command { dup_states $which_state }
	.b2 add command -label "Rename state" \
		-command { edit_state $which_state }
	.b2 add command -label "Add Transition to Existing State" \
		-command { connect_state $which_state $which_x $which_y 0 }
	.b2 add command -label "Add Transition to New State" \
		-command { connect_state $which_state $which_x $which_y 1 }
	.b2 add command -label "Delete State"  \
		-command { delete_state $which_state }

	menu .b3 -tearoff 0
	.b3 add command -label "Delete" \
		-command { set curproc $belongs($which_state); \
			set ev [find_label $which_state $next_state]; \
			delete_transition $which_state $next_state $curproc $ev }
	.b3 add command -label "Edit label"   \
		-command { rename_transition $which_state $next_state }

	# shortcuts
	bind . <Control-Key-s> {do_save 0}
	bind . <Control-Key-a> {incr cnt; new_state $cnt}	;# +state
	bind . <Control-Key-c> {add_data}	;# +context
	bind . <Control-Key-i> {do_sim}		;# +sim
	bind . <Control-Key-v> {do_check}	;# +verify
	bind . <Control-Key-m> {do_pml 2}	;# +model
	bind . <Control-Key-r> {reset_states}	;# +reset

	bind . <Key-Up>   {}
	bind . <Key-Down> {}

	.b.notes.t insert end "Shortcuts:\n"
	.b.notes.t insert end "  Save:     Ctl-s\n"
	.b.notes.t insert end "  +state:   Ctl-a\n"
	.b.notes.t insert end "  +context: Ctl-c\n"
	.b.notes.t insert end "  +sim:     Ctl-i\n"
	.b.notes.t insert end "  +verify:  Ctl-v\n"
	.b.notes.t insert end "  +model:   Ctl-m\n"
	.b.notes.t insert end "  +reset:   Ctl-r\n"

set declarations ""

proc do_help {} {
	catch { destroy .help }
	toplevel .help -bg honeydew
	wm title .help "Help"
	wm geometry .help +100+100

	text .help.t
	.help.t insert end "
+state will place a new state on the canvas
	you can change its name, or assign it to a process,
	with the button-3 menu
	add outgoing transitions for this state with the
	transition option in this menu

	if you start a transition label with the character '!'
	or '?' this is interpreted to be output to ('!')
	or input from ('?') a predefined channel named q
	you can, however, also define your own channels in
	the context segment and use standard Promela syntax
	to send and receive messages to these

	to let a transition print a string or value in the
	log-window (on the left) use the predefined function
	'show', as in:
		show(\"anything %d\\n\", d)
	if you start the string to be printed with a tilde ~
	then it will be highlighted in the log-window

	transition labels should not contain the
	reserved characters: # or @
	if needed though, you can define a macro name
	and expand it in the +context specification

+context allows you to define a simple context for the
	automata you've defined -- you can use this to
	specify variable declarations, define inlines

	you can also define ltl requirements here, or
	define additional processes in Promela

	you can add local variable declarations for a
	process defined as an automaton on the canvas
	with macro-definitions, using the process-name
	prefixed with the string \"local_\".
	For instance:
	 #define local_threadname  int x\[5\]; bool y;

+model	generates a spin model and places it in a file
	with extension .pml.

+sim	will generate a spin model and perform either a
	random simulation, or playback of an error-trail
	found with the verification option

+verify	perform a formal verification

+reset	reset states and transitions highlighted in an
	earlier simulation

Menus for manipulating states or transitions show up
if you click mouse buttons 1 or 2. Move states around
with mouse button 1 down. Moving the initial state of
a process moves all connected states.

Shortcuts:
	Save:     Ctl-s
	+state:   Ctl-a
	+context: Ctl-c
	+sim:     Ctl-i
	+verify:  Ctl-v
	+model:   Ctl-m
	+reset:   Ctl-r
"
	frame .help.bot -bg ivory
	 button .help.bot.close -text "Close" \
		-bg white -command { destroy .help }
	 pack append .help.bot \
		.help.bot.close { right }

	pack append .help \
		.help.t { top fillx filly expand } \
		.help.bot { top fillx }
}

proc add_data {} {
	global declarations anychange

	catch { destroy .data }
	toplevel .data -bg honeydew
	wm title .data "Data Declarations"
	wm geometry .data +100+100

	text .data.t
	if {$declarations == ""} {
		.data.t insert end "// optionally: add data, channels,\n"
		.data.t insert end "// and/or proctype declarations here"
	}
	.data.t insert end "$declarations"

	frame .data.bot -bg ivory
	  button .data.bot.accept -text "Set" \
		-bg white \
		-command {incr anychange; set declarations [.data.t get 0.0 end]; destroy .data}
	  button .data.bot.reject -text "Cancel" \
		-bg white \
		-command { destroy .data }
	  pack append .data.bot \
		.data.bot.accept {right} \
		.data.bot.reject {right}

	pack append .data \
		.data.t { top fillx filly expand } \
		.data.bot { top fillx }
}

set prompt 1

proc rename_transition {from to} {
	global which_x which_y

	set ev [find_label $from $to]

	set ww [expr [string length $ev] * 10]
	if {$ww < 200} { set ww 200 }

	catch { destroy .lab }
	toplevel .lab -bg honeydew
	wm geometry .lab ${ww}x50+$which_x+$which_y

	new_row "lab" "top" "Label" 10

	frame .lab.bot -bg white
	button .lab.bot.accept -text "Set" \
		-bg white \
		-command "do_rename $from $to"
	button .lab.bot.reject -text "Cancel" \
		-bg white \
		-command { destroy .lab }

	pack append .lab.bot \
		.lab.bot.accept {right} \
		.lab.bot.reject {right}
	pack append .lab \
		.lab.bot { right }

	.lab.top.ent insert end "$ev"

	bind .lab.top.ent <Return> "do_rename $from $to"
}

proc do_rename {from to} {
	global cv inp out statename anychange

	set ev [.lab.top.ent get]

	if [bad_label "$ev"] {
		destroy .lab
		return
	}
	$cv itemconfigure lab$from$to -text "$ev"

	# all inputs of $to

	set nt $statename($from)
	if {[info exists inp($to)]} {
		set n [llength $inp($to)]
		for {set i 0} {$i < $n} {incr i} {
			set tr [lindex $inp($to) $i]
			set pr [lindex $tr 0]
			if {$nt == [lindex $tr 1]} {
				set inp($to) [lreplace $inp($to) $i $i [list $pr $nt "$ev"]]
				break
	}	}	}

	# all outputs of $from
	set fr $statename($to)
	if {[info exists out($from)]} {
		set n [llength $out($from)]
		for {set i 0} {$i < $n} {incr i} {
			set tr [lindex $out($from) $i]
			set pr [lindex $tr 0]
			if {$fr == [lindex $tr 1]} {
				set out($from) [lreplace $out($from) $i $i [list $pr $fr "$ev"]]
				break
	}	}	}
	incr anychange

	catch { destroy .lab }
}

proc start_over {} {
	global Fname prompt cnt state_exists declarations
	global pnames processes anychange

	if {$anychange > 0} {
		set answer [tk_messageBox -type yesno -icon question \
			-message "There are unsaved changes, discard these?"]
		switch -- $answer {
		yes	{ set anychange 0 }
		no	{ return }
	}	}

#	if {$cnt > 0} {
#		set answer [tk_messageBox -type yesno -icon question \
#			-message "Delete all current states?"]
#		switch -- $answer {
#		yes	{ }
#		no	{ return }
#	}	}

	set Fname ""
	set prompt 0
	for {set i 0} {$i <= $cnt} {incr i} {
		if {[info exists state_exists(S$i)]} {
			delete_state $i
	}	}

	for {set i 0} {$i < [llength $pnames]} {incr i} {
		set x [lindex $pnames $i]
		set processes($x) {}
	}
	set pnames {}
	set prompt 1
	set declarations ""
	set anychange 0
}

proc align_grid {} {
	global cv cnt state_exists

	set G 6	;# grid spacing
	for {set i 0} {$i <= $cnt} {incr i} {
		if {[info exists state_exists(S$i)]} {
			set op [$cv coords w$i]
			set  x [lindex $op 0]
			set  y [lindex $op 1]
			set  x [expr {int($x)}]
			set  y [expr {int($y)}]
			set nx [expr {int($x)}]
			set ny [expr {int($y)}]
			if {($nx % $G) != 0} {
				set nx [expr $nx + $G - ($nx % $G)]
			}
			if {($ny % $G) != 0} {
				set ny [expr $ny + $G - ($ny % $G)]
			}
			if {$nx != $x || $ny != $y} {
				move_state $i $nx $ny 0
	}	}	}
}

proc find_label {from to} {
	global cnt state_exists out statename

	if {[info exists state_exists(S$from)] \
	&&  [info exists out($from)]} {
		set n [llength $out($from)]
		for {set i 0} {$i < $n} {incr i} {
			set tr [lindex $out($from) $i]
			set xn $statename($to)
			if {$xn == [lindex $tr 1]} {
				return [lindex $tr 2]
	}	}	}
	tk_messageBox -icon info -message "cannot find transition from $from to $to"
	return "unknown"
}

proc move_state {n dx dy recursive} {
	global cv ShW ShH out inp reachable with_shadows anychange
	global processes curproc belongs statename shadow_dist

	set mx [$cv canvasx $dx]
	set my [$cv canvasy $dy]

	if {$recursive} {
		set op [$cv coords w$n]
		set deltax [expr $dx - [lindex $op 0]]
		set deltay [expr $dy - [lindex $op 1]]
	}
	$cv coords s$n \
		[expr $mx - $ShW/2] [expr $my - $ShH/2] \
		[expr $mx + $ShW/2] [expr $my + $ShH/2]

	if {$with_shadows} {
		$cv coords shadow$n \
			[expr $mx - $ShW/2+$shadow_dist] [expr $my - $ShH/2+$shadow_dist] \
			[expr $mx + $ShW/2+$shadow_dist] [expr $my + $ShH/2+$shadow_dist]
	}

	$cv coords w$n $mx $my

	set sn $statename($n)
	if {[info exists out($n)]} {
		set nn [llength $out($n)]
		for {set i 0} {$i < $nn} {incr i} {
			set tr [lindex $out($n) $i]
			if {[llength $tr] == 3} {
				set curproc [lindex $tr 0]
				set dst [lindex $tr 1]
				set evt [lindex $tr 2]
				reconnect $sn $dst $evt
	}	}	}
	if {[info exists inp($n)]} {
		set nn [llength $inp($n)]
		for {set i 0} {$i < $nn} {incr i} {
			set tr [lindex $inp($n) $i]
			if {[llength $tr] == 3} {
				set curproc [lindex $tr 0]
				set src [lindex $tr 1]
				set evt [lindex $tr 2]
				reconnect $src $sn $evt
	}	}	}

	if {$recursive && $processes($belongs($n)) == $n} {	;# initial state
		set reachable {}
		collect_reachable $belongs($n) $n
		set reachable [lreplace $reachable 0 0]
		for {set i 0} { $i < [llength $reachable]} {incr i} {
			set c [lindex $reachable $i]
			set op [$cv coords w$c]
			set ox [lindex $op 0]
			set oy [lindex $op 1]
			move_state $c [expr $ox + $deltax] [expr $oy + $deltay] 0
	}	}
	set anychange 1
}

proc dup_states {n} {
	global reachable belongs cnt processes
	global cv statename out curproc anychange

	set dx 300
	set dy 0

	set reachable {}
	collect_reachable $belongs($n) $n
	set ln [llength $reachable]
	set curproc $belongs($n)
	set ois $processes($curproc)	;# old initial state
	set nis -1			;# new initial state
	set curproc ${curproc}_
	for {set i 0} {$i < $ln} {incr i} {
		set tocopy [lindex $reachable $i]
		set cl [$cv coords w$tocopy]
		set clx [lindex $cl 0]
		set cly [lindex $cl 1]

		incr cnt
		new_state $cnt
		set snm $statename($tocopy)
		set statename($cnt) ${snm}_
		$cv itemconfigure w$cnt -text "$statename($cnt)"
		move_state $cnt [expr $clx + $dx] [expr $cly + $dy] 0
		if {$tocopy == $ois} {
			set nis $cnt
		}
	}
	for {set i 0} {$i < $ln} {incr i} {
		set tocopy [lindex $reachable $i]
		set snm $statename($tocopy)
		if {![info exists out($tocopy)]} {
			continue
		}
		for {set j 0} {$j < [llength $out($tocopy)]} {incr j} {
			set tgt [lindex $out($tocopy) $j]
			set tr [lindex $tgt 1]
			set ev [lindex $tgt 2]
			add_transition ${snm}_ ${tr}_ $ev
	}	}
	if {$nis != -1} {
		add_process $curproc $nis
	}
	incr anychange
}

proc delete_state {n} {
	global cv state_exists inp out processes
	global statename belongs prompt anychange

	set sn $statename($n)

	if {$prompt} {
		set answer [tk_messageBox -type yesno -icon question \
			-message "Delete state $sn?"]
		switch -- $answer {
		yes	{ }
		no	{ return }
	}	}
	catch { $cv delete s$n }
	catch { $cv delete shadow$n }
	catch { $cv delete w$n }
	catch { unset state_exists(S$n) }

	if {[info exists inp($n)]} {
		set to $n
		set nn [llength $inp($n)]
		for {set i 0} {$i < $nn} {incr i} {
			set tr   [lindex $inp($n) $i]
			set curproc [lindex $tr 0]
			set lbl     [lindex $tr 2]
			set from [find_state $curproc [lindex $tr 1] 0]
			$cv delete con$from$to
			$cv delete lab$from$to

			# delete corresponding entry from out($from)
			if {[info exists out($from)]} {
				set j [lsearch $out($from) [list $curproc $sn $lbl]]
				set out($from) [lreplace $out($from) $j $j]
		}	}
		unset inp($n)
	}
	if {[info exists out($n)]} {
		set from $n
		set nn [llength $out($n)]
		for {set i 0} {$i < $nn} {incr i} {
			set tr [lindex $out($n) $i]
			set curproc [lindex $tr 0]
			set lbl     [lindex $tr 2]
			set to [find_state $curproc [lindex $tr 1] 0]
			$cv delete con$from$to
			$cv delete lab$from$to
			# delete entry $from from inp($to)
			if {[info exists inp($to)]} {
				set xn $statename($from)
				set j [lsearch $inp($to) [list $curproc $xn $lbl]]
				set inp($to) [lreplace $inp($to) $j $j]
		}	}
		unset out($n)
	}
	if {$processes($belongs($n)) == $n} {
		set processes($belongs($n)) -1
		unset belongs($n)
	}
	catch { unset statename($n) }
	incr anychange
}

proc fillin_p {n} {
	global belongs
# puts "Here $n $belongs($n)"
	return $belongs($n)
}
proc fillin_x {n} {
	global cv ShW
	set c [$cv coords s$n]
	return [expr [lindex $c 0] - $ShW/2]
}
proc fillin_y {n} {
	global cv
	set c [$cv coords s$n]
	return [lindex $c 1]
}

proc new_state {nr} {
	global cv SW SH ShW ShH  WW belongs curproc processes
	global SCX SCY CH state_exists cnt Helv statename anychange
	global with_shadows shadow_col state_shape shadow_dist

	set ocnt $cnt
	set cnt $nr

	if {[info exists state_exists(S$nr)]} {
		tk_messageBox -icon info -message "error: state $nr exists"
		set cnt $ocnt
		return
	}

	set statename($nr) "S$nr"
	set belongs($nr) $curproc
	if {![info exists processes($curproc)]} {
		set processes($curproc) 0
	}

	if {$with_shadows} {
	$cv create $state_shape \
		[expr $SCX - $ShW/2+$shadow_dist] [expr $SCY - $ShH/2+$shadow_dist] \
		[expr $SCX + $ShW/2+$shadow_dist] [expr $SCY + $ShH/2+$shadow_dist] \
		-outline $shadow_col -fill $shadow_col -tags shadow$nr
	}
	$cv create $state_shape \
		[expr $SCX - $ShW/2] [expr $SCY - $ShH/2] \
		[expr $SCX + $ShW/2] [expr $SCY + $ShH/2] \
		-outline royalblue -fill azure -tags s$nr

	$cv create text $SCX $SCY \
		-font $Helv \
		-text "$statename($nr)" \
		-tags w$nr

	$cv bind s$nr <Any-Enter> "$cv create text \
		\[fillin_x $nr\] \[fillin_y $nr] \
		-text \[fillin_p $nr\] -font $Helv -fill red -tags goo "
	$cv bind s$nr <Any-Leave> "$cv delete goo"
	$cv bind s$nr <B1-Motion>     "move_state    $nr %x %y 1"

	$cv bind s$nr <ButtonPress-3> "state_menu $nr %X %Y"
	$cv bind w$nr <ButtonPress-3> "state_menu $nr %X %Y"

	$cv bind s$nr <ButtonPress-2> "state_menu $nr %X %Y"
	$cv bind w$nr <ButtonPress-2> "state_menu $nr %X %Y"

	set state_exists(S$nr) 1

	incr SCY 10
#	incr SCX 10

	set cnt $ocnt
	incr anychange
}

proc reset_states {} {
	global cnt cv state_exists highlights
	global SCX SCY

	set SCY 60
	set SCX 60

	for {set c 0} {$c <= $cnt} {incr c} {
		if {![info exists state_exists(S$c)]} {
			continue
		}
		$cv itemconfigure s$c -fill azure
		$cv itemconfigure w$c -fill black
		update
	}

	for {set c 0} {$c < [llength $highlights]} {incr c} {
		set t [lindex $highlights $c]
		$cv itemconfigure $t -width 1 -fill royalblue
	}
	set highlights {}
	.b.notes.t delete 1.0 end
}

proc state_menu {nr x y} {
	global which_state which_x which_y
	set which_state $nr
	set which_x $x
	set which_y $y
	tk_popup .b2 $x $y
}

proc trans_menu {src ns x y} {
	global which_state next_state which_x which_y

	set which_state $src
	set next_state $ns
	set which_x $x
	set which_y $y
	tk_popup .b3 $x $y
}

proc new_row {base nm s lw} {

	frame .$base.$nm
	  label .$base.$nm.lb -width $lw \
		-text "${s}:" \
		-bg honeydew -fg black \
		-justify left
	  entry .$base.$nm.ent -width 45 \
		-relief flat \
		-bg white \
		-justify center
	  pack append .$base.$nm \
		.$base.$nm.lb {left} \
		.$base.$nm.ent {left fillx}
	pack append .$base .$base.$nm { top }
}

proc delete_transition {from to cp ev} {
	global cv out inp statename anychange

	set xn $statename($from)
	set yn $statename($to)

	set answer [tk_messageBox -type yesno -icon question \
		-message "Delete transition from $xn to $yn ?"]

	switch -- $answer {
	no	{ return }
	yes	{ }
	}
	catch " $cv delete con$from$to "
	catch " $cv delete lab$from$to "
	set nto $statename($to)
	set i [lsearch $out($from) [list $cp $nto "$ev"] ]
	catch {
		if {$i < 0} {
			tk_messageBox -icon info \
			-message "error: No such transition -- $cp $to $ev"
		} else {
			set out($from) [lreplace $out($from) $i $i]
		}
	}
		  set nfr $statename($from)
	catch {
		set i [lsearch $inp($to) [list $cp $nfr "$ev"] ]
		if {$i < 0} {
			tk_messageBox -icon info \
			-message "error: No such transition -- $cp $to $ev"
		} else {
			set inp($to) [lreplace $inp($to) $i $i]
	}	}
	incr anychange
}

proc reconnect {fr to ev} {
	global cv curproc Helv

	set HSH 26.0	;# horizontal shift of transition label
	set VSH 12.0	;# vertical shift

	set src [find_state $curproc $fr 0]
	if {$src < 0} {
		return
	}
	set ns [find_state $curproc $to 0]
	if {$ns < 0} {
		return
	}
	set xx [$cv coords s$src]
		set f_ulx [lindex $xx 0]
		set f_uly [lindex $xx 1]
		set f_lrx [lindex $xx 2]
		set f_lry [lindex $xx 3]
		set f_mx  [expr $f_ulx + ($f_lrx - $f_ulx)/2]
		set f_my  [expr $f_uly + ($f_lry - $f_uly)/2]

	set yy [$cv coords s$ns]
		set t_ulx [lindex $yy 0]
		set t_uly [lindex $yy 1]
		set t_lrx [lindex $yy 2]
		set t_lry [lindex $yy 3]
		set t_mx  [expr $t_ulx + ($t_lrx - $t_ulx)/2]
		set t_my  [expr $t_uly + ($t_lry - $t_uly)/2]

	set mx [expr $f_mx + ($t_mx - $f_mx)/2]
	set my [expr $f_my + ($t_my - $f_my)/2]

	set tx [expr $f_mx + ($t_mx - $f_mx)/3]
	set ty [expr $f_my + ($t_my - $f_my)/3]

	set lx $tx
	set ly $ty

	if {[expr abs($mx - $tx)] < 5} {	;# vertical
		if {$my > $ty} {
		#	set mx [expr $mx - $HSH]
			set lx [expr $lx - $HSH]
		} else {
		#	set mx [expr $mx + $HSH]
			set lx [expr $lx + $HSH]
		}
	} else {
		if {[expr abs($my - $ty)] < 5} { ;# horizontal
			if {$mx > $tx} {
		#		set my [expr $my - $VSH]
				set ly [expr $ly - $VSH]
			} else {
		#		set my [expr $my + $VSH]
				set ly [expr $ly + $VSH]
			}
		} else {
			if {$my > $ty} {
		#		set mx [expr $mx - $HSH]
				set lx [expr $lx - $HSH]
			} else {
		#		set mx [expr $mx + $HSH]
				set lx [expr $lx + $HSH]
	}	}	}

	catch " $cv delete con$src$ns "
	catch " $cv delete lab$src$ns "
  if {1} {
	$cv create line $f_mx $f_my  $mx $my \
		-fill royalblue -arrow last \
		-width 1 -tags con$src$ns
	$cv create line $mx $my  $t_mx $t_my \
		-fill royalblue \
		-width 1 -tags con$src$ns
  } else {
	$cv create line $f_mx $f_my  $mx $my $t_mx $t_my\
		-fill royalblue -arrow last -smooth 1 \
		-width 1 -tags con$src$ns
  }
	$cv create text $lx $ly -font $Helv -text "$ev" -tags lab$src$ns

	# help mac users find the button...
	$cv bind lab$src$ns <ButtonPress-1> "trans_menu $src $ns %X %Y"
	$cv bind lab$src$ns <ButtonPress-2> "trans_menu $src $ns %X %Y"
	$cv bind lab$src$ns <ButtonPress-3> "trans_menu $src $ns %X %Y"

	$cv lower con$src$ns
}

proc add_transition {fr to ev} {
	global out inp curproc anychange

	if {$ev == ""} { set ev "true" }

	set a [find_state $curproc $fr 0]
	set b [find_state $curproc $to 0]

	if {$a == -1 || $b == -1} { return }

	if {[info exists out($a)]} {
		# make sure that there is not already
		# a transition from a to b
		for {set i 0} {$i < [llength $out($a)]} {incr i} {
			set s [lindex $out($a) $i]
			set s [lindex $s 1]
			if {$s == $to} {
				tk_messageBox -icon info \
					-message "There already is a transition from $fr to $to"
				return
	}	}	}

	if {![info exists out($a)] \
	|| [lsearch $out($a) [list $curproc $to $ev] ] < 0} {
		lappend out($a) [list $curproc $to $ev]
	}
	if {![info exists inp($b)] \
	|| [lsearch $inp($b) [list $curproc $fr $ev] ] < 0} {
		lappend inp($b) [list $curproc $fr $ev]
	}
	reconnect $fr $to $ev
	incr anychange
}

set pnames    {}
set reachable {}

proc collect_reachable {pnm n} {
	global reachable out

	if {[lsearch $reachable $n] >= 0} {
		return
	}
	set reachable [lappend reachable $n]
	if {[info exists out($n)]} {
		for {set i 0} {$i < [llength $out($n)]} {incr i} {
			set s [lindex $out($n) $i]
			set s [lindex $s 1]
			set s [find_state $pnm $s 0]
			collect_reachable $pnm $s
	}	}
}

proc add_process {pnm n} {
	global curproc cv		;# scalars
	global processes belongs	;# arrays
	global pnames reachable		;# lists
	global inp out anychange

	# define n as the initial state of process pnm
	# mark all states reachable from this state
	# and check that there's no incoming transition from any other state to n

	set reachable {}
	if {[info exists belongs($n)]} {	;# else it's a new state/process
		collect_reachable $belongs($n) $n	;# old pnm n
	
		if {[info exists inp($n)]} {
			for {set i 0} {$i < [llength $inp($n)]} {incr i} {
				set s [lindex $inp($n) $i]
				set ns [lindex $s 1]
				set ns [find_state $belongs($n) $ns 0]
				if {[lsearch $reachable $ns] < 0} {
					tk_messageBox -icon info -message \
					"error: input-state $ns is not reachable from $n :: $reachable"
					return
		}	}	}

		# check if n is currently set as initial under a different processname
	        catch {
		  set p_old $belongs($n)	;# n is currently in process p_old
		  set i_old $processes($p_old)	;# which has i_old as an initial state
		  if {$i_old == $n} {
		 	set processes($p_old) -1
		  } else {
			set i_op $processes($pnm)
			if {$i_op != $n} {
				$cv itemconfigure s$i_op -width 1	;# unmark
		  }	}
	}	}

	# mark all states in $reachable as belonging to $pnm
	for {set i 0} {$i < [llength $reachable]} {incr i} {
		set x [lindex $reachable $i]
		set belongs($x) $pnm
		# update inp and out entries
		if {[info exists inp($x)]} {
# puts "update inp for $x -- old: $inp($x) length: [llength $inp($x)]"
			set newlist {}
			for {set j 0} {$j < [llength $inp($x)]} {incr j} {
				set z [lindex $inp($x) $j]	;# triple, first entry is pnm
# puts "	retrieved $z len: [llength $z]"
				set z0 [lindex $z 0]	;# procname
				set z1 [lindex $z 1]	;# src state
				set z2 [lindex $z 2]	;# event
				lappend newlist [list $pnm $z1 $z2]
			}
			set inp($x) $newlist
# puts "new: $inp($x)"
		}
		if {[info exists out($x)]} {
# puts "update out for $x -- old: $out($x)"
			set newlist {}
			for {set j 0} {$j < [llength $out($x)]} {incr j} {
				set z [lindex $out($x) $j]	;# triple, first entry is pnm
				set z0 [lindex $z 0]	;# procname
				set z1 [lindex $z 1]	;# dst state
				set z2 [lindex $z 2]	;# event
				lappend newlist [list $pnm $z1 $z2]
			}
			set out($x) $newlist
# puts "new: $out($x)"
	}	}

	# remove the old name
	# but only if all the old initial state $processes($curproc)
	# is reachable from the new one

	if {$curproc != "---" && $processes($curproc) != {} \
	&&  [lsearch $reachable $processes($curproc)] >= 0} {
		set x [lsearch $pnames $curproc]	;# the old process name
		if {$x >= 0} {
			set pnames [lreplace $pnames $x $x]
	}	}

	set curproc $pnm		;# the new process name
	set processes($pnm) $n		;# the new initial state
	$cv itemconfigure s$n -width 2

	if {[lsearch $pnames $pnm] < 0} {
		lappend pnames $pnm
	}
# puts "process $pnm istate $n"

# puts "pnames: $pnames"

	incr anychange
	return
}

proc go_ahead {} {
	global which_state

	set pnm [.process.pnm.ent get]
	add_process $pnm $which_state
	destroy .process 
}

proc new_state_name {on nm} {
	global inp out state_exists cnt anychange

	for {set i 0} {$i <= $cnt} {incr i} {
		if {![info exists state_exists(S$i)]} {
			continue
		}
		if {[info exists inp($i)]} {
			set ln [llength $inp($i)]
			for {set j 0} {$j < $ln} {incr j} {
				set tr [lindex $inp($i) $j]
				if {[lindex $tr 1] != $on} {
					continue
				}
				set p0 [lindex $tr 0]
				set ev [lindex $tr 2]
				set ne [list $p0 $nm "$ev"]
				set inp($i) [lreplace $inp($i) $j $j $ne]
				break
		}	}

		if {[info exists out($i)]} {
			set ln [llength $out($i)]
			for {set j 0} {$j < $ln} {incr j} {
				set tr [lindex $out($i) $j]
				if {[lindex $tr 1] != $on} {
					continue
				}
				set p0 [lindex $tr 0]
				set ev [lindex $tr 2]
				set ne [list $p0 $nm "$ev"]
				set out($i) [lreplace $out($i) $j $j $ne]
				break
	}	}	}
	incr anychange
}

proc setup_rename {} {
	global cv which_state statename belongs
	global nowarn anychange

	set on $statename($which_state)
	set sn [.rename.rnm.ent get]

	set curproc $belongs($which_state)
	set nowarn 1
	set e [find_state $curproc $sn 0]
	set nowarn 0
	if {$e != -1} {
		tk_messageBox -icon info -message "error: state $sn already exists"
		return
	}

	$cv itemconfigure w$which_state -text "$sn"
	set statename($which_state) "$sn"

	catch { destroy .rename }
	new_state_name $on $sn
	incr anychange
}

proc edit_state {n} {
	global which_state which_x which_y statename

	catch { destroy .rename }
	toplevel .rename -bg honeydew
	wm title .rename "Edit Statename"
	wm geometry .rename 180x50+$which_x+$which_y

	new_row "rename" "rnm" "Name" 10
	bind .rename.rnm.ent <Return> { setup_rename }
	.rename.rnm.ent insert end $statename($n)

	frame .rename.b -bg white
		button .rename.b.accept \
			-text "Set" -bg white \
			-command { setup_rename }
		button .rename.b.reject -text "Cancel" \
			-command { destroy .rename } -bg white
	pack append .rename.b \
		.rename.b.accept {right} \
		.rename.b.reject {right }
	pack append .rename \
		.rename.b {right}
}

proc set_initial {n} {
	global which_state curproc belongs which_x which_y
	# define a name and an initial state for a process

	set which_state $n

	catch { destroy .process }
	toplevel .process -bg honeydew
	wm title .process "Define Process"
	wm geometry .process 180x50+$which_x+$which_y

	new_row "process" "pnm" "         Name" 10
	bind .process.pnm.ent <Return> { go_ahead }

	set curproc "$belongs($n)"
	.process.pnm.ent insert end "$curproc"

	frame .process.b -bg white
		button .process.b.accept \
			-text "Set" -bg white \
			-command { go_ahead }
		button .process.b.reject -text "Cancel" \
			-command { destroy .process } -bg white
	pack append .process.b \
		.process.b.accept {right} \
		.process.b.reject {right }
	pack append .process \
		.process.b {right}
}

proc bad_label {ev} {

	if {[string first "#" $ev] >= 0 \
	||  [string first "@" $ev] >= 0} {
		tk_messageBox -icon info \
			-message "error: do not use '#' or '@' in transition labels"
		return 1
	}
	return 0
}

proc do_trans {nr} {

	set fr [.query.from.ent  get]
	set to [.query.dest.ent  get]
	set ev [.query.event.ent get]

	if {![bad_label "$ev"]} {
		add_transition $fr $to "$ev"
	}
}

proc connect_state {nr x y newstate} {
	global cv SW ShH SH curproc belongs statename cnt

	incr x $SW

	catch { destroy .query }
	toplevel .query -bg honeydew
	wm title .query "Transition"
	wm geometry .query 180x110+$x+$y

	new_row "query" "from"	"   From" 7
	new_row "query" "dest"  "     To" 7
	new_row "query" "event" "  Event" 7

	frame .query.b -bg white
		button .query.b.accept \
			-bg white -text "Set" \
			-command " set curproc $belongs($nr); do_trans $nr; destroy .query "
		button .query.b.reject -text "Cancel" \
			-command { destroy .query } -bg white
	pack append .query.b \
		.query.b.accept {right} \
		.query.b.reject {right }
	pack append .query \
		.query.b {right}

	.query.from.ent insert end $statename($nr)

	set curproc $belongs($nr)

	if {$newstate} {
		incr cnt
		new_state $cnt
		set op [$cv coords w$nr]
		set ox [lindex $op 0]
		set oy [lindex $op 1]
		move_state $cnt $ox [expr $oy + $SH] 0

		.query.dest.ent insert end $statename($cnt)
	}
	bind .query.dest.ent <Return> {
		find_state $curproc [.query.dest.ent get] 1
	}
	bind .query.event.ent <Return> " \
		 set curproc $belongs($nr); do_trans $nr; destroy .query "
}

proc do_open {} {
	global Fname anychange

	if {$anychange > 0} {
		set answer [tk_messageBox -type yesno -icon question \
			-message "There are unsaved changes, discard these?"]
		switch -- $answer {
		yes	{ set anychange 0 }
		no	{ return }
	}	}

	set types { {{Tau Files} {.tau}} {{All Files} *} }

	set file [tk_getOpenFile -defaultextension .tau -filetypes $types]
	if {$file != ""} {
		if { [openfile $file] } {
			set Fname $file
		}
	}
}

proc mk_state {nr} {
	global state_exists

	if {![info exists state_exists(S$nr)]} {
		new_state $nr
	}
}

proc do_defers {} {
	global defer curproc
	global cv processes

	if {[llength $defer] > 0} {
		$cv itemconfigure s$processes($curproc) -width 2
	}
	for {set i 0} {$i < [llength $defer]} {incr i} {
		set de [lindex $defer $i]
		set de_sn [lindex $de 0]
		set de_tg [lindex $de 1]
		set de_ev [lindex $de 2]
# puts "	trans $de ($curproc)"
		add_transition $de_sn $de_tg "$de_ev"
	}
	set defer {}
}

proc openfile {fn} {
	global cv state_exists cnt ShW ShH statename shadow_dist
	global curproc which_state defer declarations with_shadows
	global anychange

	if {$anychange > 0} {
		set answer [tk_messageBox -type yesno -icon question \
			-message "There are unsaved changes, discard these?"]
		switch -- $answer {
		yes	{ set anychange 0 }
		no	{ return }
	}	}

	if [catch {set fd [open $fn r]} errmsg] {
		tk_messageBox -icon info -message "$errmsg"
		return 0
	}

	start_over

	set mcnt 0
	set defer {}
	while {[gets $fd line] > -1} {
# puts $line
		if {[string length $line] == 0 || \
		    [string first "#" $line] == 0} {
# puts "comment"
			continue
		}
		if {[string first "data" $line] == 0} {
			while {[gets $fd line] > -1} {
				if {[string first "atad" $line] == 0} {
					break
				}
				set declarations "${declarations}$line\n"
			}
# puts "data"
			continue
		}
		set f [split $line "@"]
		set y [llength $f]
# puts "here $y"
		switch -- $y {
		3 {	;# trans,event,state
			set ev [lindex $f 1]
			set tg [lindex $f 2]
			lappend defer [list $sn $tg "$ev"]
		  }
		4 {	;# process,name,is,0
			do_defers
			set pn [lindex $f 1]
			set is [lindex $f 2]
# puts "add_proc $pn $is"
			set curproc "---"
			add_process $pn $is
			set curproc $pn
		  }
		5 {	;# state,nr,nm,x,y
			set nr [lindex $f 1]
			set sn [lindex $f 2]
			set ux [lindex $f 3]
			set uy [lindex $f 4]
			set ux [expr 5 * {round ($ux/5)}]
			set uy [expr 5 * {round ($uy/5)}]
# puts "	mk_state $pn $nr $sn"
			mk_state $nr
			$cv coords s$nr \
				[expr $ux - $ShW/2] [expr $uy - $ShH/2] \
				[expr $ux + $ShW/2] [expr $uy + $ShH/2]
			$cv coords w$nr $ux $uy
			if {$with_shadows} {
				$cv coords shadow$nr \
					[expr $ux - $ShW/2+$shadow_dist] [expr $uy - $ShH/2+$shadow_dist] \
					[expr $ux + $ShW/2+$shadow_dist] [expr $uy + $ShH/2+$shadow_dist]
			}
			$cv itemconfigure w$nr -text "$sn"
			set statename($nr) "$sn"
			if {$nr > $cnt} {
				set cnt $nr
		  }	}
		default {
			tk_messageBox -icon info -message "bad format ($y): $line -- ignored"
	}	} }
	do_defers
# puts "done with openfile"

	close $fd

	wm title . "TauTool - $fn"

	set anychange 0

	return 1
}

proc do_exit {} {
	global anychange fname

	if {$anychange > 0} {
		set answer [tk_messageBox -type yesno -icon question \
			-message "There are unsaved changes, discard these?"]
		switch -- $answer {
		yes	{ set anychange 0 }
		no	{ return }
	}	}

	catch { eval exec "rm -f _tmp2_ pan _spin_nvr.tmp ${fname}.pml ${fname}.trail" }
	destroy .
	exit 0
}

proc do_save {newname} {
	global Fname anychange

	if {$newname || $Fname == ""} {
		set fileselect "tk_getSaveFile"
		# get the file (return if the file selection dialog canceled)
		switch -- [set file [eval $fileselect]] "" return
		set Fname $file
		wm title . "TauTool - $Fname"
	}

	set d [string last "." $Fname]

	if {$d > 0} {
		set f [string range $Fname 0 $d]
	} else {
		set f $Fname
	}
	set Fname ${f}tau

	if {[save $Fname] == 1} {
		set anychange 0
	}
}

proc flash_state {c} {
	global cv

	$cv itemconfigure s$c -fill red
	update
	after 50
	$cv itemconfigure s$c -fill azure
	update
}

proc find_state {pn nm flash} {
	global cv cnt ulx uly state_exists belongs statename
	global nowarn

	set elsewhere -1
	for {set c 0} {$c <= $cnt} {incr c} {
		if {![info exists state_exists(S$c)]} {
			continue
		}
		set snm $statename($c)
		set wp  $belongs($c)
# puts "see $snm want $nm in $pn"
		if {$snm == $nm} {
			if {$pn == $wp} {
				set z [$cv coords s$c]
				set ulx [lindex $z 0]
				set uly [lindex $z 1]
				if {$flash} {
					flash_state $c
				}
				return $c
			} else {
				set elsewhere $c
	}	}	}

	if {$elsewhere >= 0} {
		set c $elsewhere
		flash_state $c
		if {$nowarn == 0} {
		  tk_messageBox -icon info \
			-message "State $nm belongs to Process $belongs($c), not $pn"
		} else {
			return 1	;# dont allow same state name anywhere
			# because out(name) doesnt contain a processname owner
		}
	} else {
		if {$nowarn == 0} {
		  tk_messageBox -icon info \
			-message "Process $pn: state $nm not found"
	}	}
	return -1
}

proc show_log {s} {
	.b.log.t insert end "$s"
	.b.log.t yview -pickplace end
}

proc save {fn} {
	global cv state_exists cnt pnames processes out
	global ulx uly declarations statename

	if [catch {set fd [open $fn w]} errmsg] {
		tk_messageBox -icon info -message "$errmsg"
		return 0
	}
	if {$declarations != ""} {
		set declarations [string trimright "$declarations" "\n"]
		puts $fd "data"
		puts $fd "$declarations"
		puts $fd "atad\n"
	}

	set cnt_saved 0
# puts "pnames: $pnames -- [info exists state_exist(S2)]"
	for {set j 0} {$j < [llength $pnames]} {incr j} {
		set want [lindex $pnames $j]

		set is $processes($want)	;# initial state
# puts "initial state $want is $is -- [info exists state_exists(S$is)] -- [info exists state_exists(S2)]"
		if {$is < 0} { continue }
		if {![info exists state_exists(S$is)]} { continue }
		lappend todo $is
		puts $fd "process@$want@$is@0"

		set visited {}
		while {[llength $todo] > 0} {
			set nm   [lindex $todo 0]
			set todo [lreplace $todo 0 0]
# puts "check $nm -- $todo -- $visited"
			if {[lsearch $visited $nm] >= 0} {
				continue
			}
			lappend visited $nm
			set z [$cv coords w$nm]
			set ulx [lindex $z 0]
			set uly [lindex $z 1]

			set snm $statename($nm)
			incr cnt_saved
			puts $fd " state@$nm@$snm@$ulx@$uly"
			if {[info exists out($nm)]} {
# puts "Out $snm -> $out($nm)"
				set n [llength $out($nm)]
				for {set i 0} {$i < $n} {incr i} {	;# successors
					set tr [lindex $out($nm) $i]
					if {$want == [lindex $tr 0]} {
# puts "tr: $want in $tr [llength $tr]"
						set dst [lindex $tr 1]
						set evt [lindex $tr 2]
						puts $fd "  trans@$evt@$dst"
# show_log "A\n"
						set xx [find_state $want $dst 0]
# show_log "B\n"
						if {$xx >= 0} {
							lappend todo $xx
					}	}
		}	}	}
		puts $fd ""
	}
	close $fd
	if {$cnt_saved == 0} {
		tk_messageBox -icon info -message "warning: no reachable states"
		return 0
	} else {
		show_log "saved $fn\n"
	}
	return 1
}

proc do_pml {how} {
	global fname Fname

	set d [string last "." $Fname]

	if {$d > 0} {
		set fname [string range $Fname 0 $d]
	} else {
		set fname $Fname
	}

	save "_tmp1_"

	if {$how == 1} {	;# for simulation/verification
		set pn "t2s -v"
	} else {		;# basic model
		set pn "t2s"
	}

	catch { eval exec "$pn _tmp1_ > ${fname}pml" } e
	catch { eval exec "rm -f _tmp1_" }

	if {$e != ""} {
		show_log "$e\n"
		catch { destroy .control }
		return ""
	} else {
		if {$how == 2} {	;# generate model
		  show_file "${fname}pml" "// File: ${fname}pml" 0
	#	  tk_messageBox -icon info -message "Wrote Spin Model into file: ${fname}pml"
		}
		return ${fname}pml
	}
}

set highlights {}
set timelag 500
set replay {}
set replay_pos 0

proc replay_up {} {
	global replay replay_pos statename curstate

	if {$replay_pos <= 0} {
		return
	}
	incr replay_pos -1

	if {[llength $replay] >= $replay_pos} {
		set x [lindex $replay $replay_pos]
		if {[llength $x] == 2} {
			set p [lindex $x 0]
			set c [lindex $x 1]

# puts "step $replay_pos of [llength $replay] $p $statename($c)"

			if {$curstate($p) == $c} {
				replay_up	;# one more up
				return
			}

			mark_state $p $c 0
	}	}
}

proc replay_down {} {
	global replay replay_pos statename curstate

	if {[expr $replay_pos + 1] >= [llength $replay]} {
		return
	}
	incr replay_pos

	set x [lindex $replay $replay_pos]
	if {[llength $x] == 2} {
		set p [lindex $x 0]
		set c [lindex $x 1]

# puts "step $replay_pos of [llength $replay] $p $statename($c)"

		if {$curstate($p) == $c} {
			replay_down	;# one more down
			return
		}
		mark_state $p $c 0
	}
}

proc mark_state {p c u} {
	global cv curstate highlights replay replay_pos

	if {[info exists curstate($p)]} {
		$cv itemconfigure s$curstate($p) -fill azure		;# old state
		$cv itemconfigure w$curstate($p) -fill black
		$cv itemconfigure con$curstate($p)$c -width 2 -fill red	;# arrow
		lappend highlights con$curstate($p)$c
	}

	$cv itemconfigure s$c -fill darkblue	;# new state
	$cv itemconfigure w$c -fill white
	set curstate($p) $c

	if {$u} {
		lappend replay "$p $c"
		incr replay_pos
	}
	update
}

proc executes {p a} {


	.b.notes.t yview -pickplace end

	if {[string first "~" $a] == 0} {
		.b.notes.t tag add highlight 0.0 0.0
		.b.notes.t tag configure highlight -background black -foreground gold
		set a [string range $a 1 end]
		.b.notes.t insert end "$a\n" highlight
	} else {
		.b.notes.t insert end "$a\n"
	}
	update
}

proc show_file { f e hunt } {
	global Helv error_trail fname sim_fd

	if {$e != ""} {
		set toshow "${e}"
	} else {
		set toshow "Result:"
	}
	set saw_error 0
	if [catch {set fd [open $f r]} e] {
		set toshow "${e}\n${toshow}\n$e"
	} else {
		set canshow [expr 1 - $hunt]
		while {[gets $fd line] > -1} {
			if {$hunt} {
				if {[string first "trail ends" $line] >= 0} {
					set canshow 1
				}
				if {$canshow} {
					set toshow "${toshow}\n${line}"
					.b.notes.t insert end "${line}\n"
					.b.notes.t yview -pickplace end
				}
				continue
			}
			set toshow "${toshow}\n${line}"
			if {[string first "errors: 1" $line]} {
				incr saw_error
		}	}
		close $fd
	}
	if {$hunt} {
		return
	}
	catch { destroy .show }
	toplevel .show -bg honeydew
	wm geometry .show +100+100

	frame .show.top
	text .show.top.t -bg ivory -yscrollcommand {.show.top.v set}
	scrollbar .show.top.v -bg honeydew -orient vert -command {.show.top.t yview }
	.show.top.t insert end "$toshow"
	pack append .show.top \
		.show.top.v { left filly } \
		.show.top.t { right fillx filly expand }

	pack append .show \
		.show.top {top fillx filly expand}

	frame .show.bot
	button .show.bot.dismiss -text "Close" \
		-bg honeydew -command { destroy .show }
	pack append .show.bot \
		.show.bot.dismiss {right}

	if {$saw_error > 0} {
		button .show.bot.replay -text "Replay Trail" \
			-bg honeydew \
			-command {
				destroy .show
				if {$sim_fd >= 0} {
					catch { close $sim_fd }
					set sim_fd -1
				}
				if {$sim_fd == -2} {
					set sim_fd -1
				}
				set error_trail ${fname}.trail
				reset_states
				do_run 1
			}
		button .show.bot.msc -text "Show Trail as MSC" \
			-bg honeydew \
			-command {
				destroy .show
				set error_trail "notempty"
				set fname [do_pml 1]
				do_run 2
			}
		pack append .show.bot \
			.show.bot.replay {right} \
			.show.bot.msc    {right}
	}

	pack append .show \
		.show.bot {top}
}

set error_trail ""

proc do_run {how} {
	global nr_steps timelag sim_fd fname
	global replay replay_pos

	# 0 - random
	# 1 - replay trail
	# 2 - trail as msc

	if {$how != 0} {
		catch { destroy .control }
		update
	} else {		;# random
		reset_states
		if {$sim_fd >= 0} {
			catch { close $sim_fd }
			set sim_fd -1
		}
#		catch { destroy .control.bot.go }
	}

	if {$how == 2} {	;# trail as msc
		show_log "spin -t -M $fname\n"
		catch "eval exec \"spin -t -M $fname &\""
		return
	}

	if {$how == 0} {
		set nr_steps [.control.top.ent get]
	} else {
		set nr_steps 10000
	}
	set replay {}
	set replay_pos 0
	bind . <Key-Up>   {}
	bind . <Key-Down> {}

	while {$nr_steps >= 1 && [sim_step] == 1} {
		set nr_steps [expr $nr_steps - 1]
		catch {
		.control.top.ent delete 0 end
		.control.top.ent insert end $nr_steps
		}
		update
		if {$timelag > 1} {
			after $timelag
		}
	}
	bind . <Key-Up>   { replay_up }		;# after trail replay
	bind . <Key-Down> { replay_down }
}

proc simulation_control {} {
	global sim_fd nr_steps fname error_trail cnt timelag

	if {$cnt == 0} { return }	;# no states to simulate

	catch { destroy .control }
	toplevel .control -bg honeydew
	wm geometry .control 320x50+50+670

	set nr_steps 100
	set sim_fd -1

	new_row "control" "top" "Max Random Steps" 16
	.control.top.ent insert end "$nr_steps"

	set fname [do_pml 1]
	if {$fname == ""} { return }

	frame .control.bot -bg ivory
		button .control.bot.close -bg white -text "Close" \
			-command { 
				catch { close $sim_fd }
				set sim_fd -2
				catch { eval exec "rm -f _tmp2_" }
				destroy .control 
			}
			pack append .control.bot \
				.control.bot.close { right }

		if {![catch {set tfd [open ${fname}.trail r]}]} {
			catch { close $tfd }
			wm geometry .control 460x50+50+670

			button .control.bot.trail -bg white \
				-text "Replay Trail" \
				-command { set error_trail ${fname}.trail; do_run 1 }

			pack append .control.bot \
				.control.bot.trail { right }

			button .control.bot.msc -bg white \
				-text "Trail as MSC" \
				-command { set error_trail ${fname}.trail; do_run 2 }

			pack append .control.bot \
				.control.bot.msc { right }
		}

		radiobutton .control.bot.fast \
			-variable timelag -value 0 -text "Fast"
		radiobutton .control.bot.slow \
			-variable timelag -value 500 -text "Slow"

		button .control.bot.go -bg white  -text "Random" \
			-command { set error_trail ""; do_run 0 }
		pack append .control.bot \
			.control.bot.go  { right }

		pack append .control.bot \
			.control.bot.fast  { right } \
			.control.bot.slow  { right }

		pack append .control \
			.control.bot { right }
}

set sim_fd -1

proc sim_step {} {
	global sim_fd nr_steps error_trail fname

	if {$sim_fd == -2} { return }
	if {$sim_fd == -1} {
		if {$fname == ""} {
			return 0
		}

		if {$error_trail == ""} {
			set rs [ expr round(1000*rand()) ]
			show_log "spin -T -u[expr 3*$nr_steps] -n$rs $fname\n"
			# visible steps are not the same as automata steps
			catch "eval exec \"spin -T -u[expr 3 * $nr_steps] -n$rs $fname > _tmp2_ \"" e
		} else {
			show_log "spin -T -t $fname\n"
			catch "eval exec \"spin -T -t $fname > _tmp2_ \"" e
		}
		if {$e != ""} {
			catch { destroy .control }
			show_file "_tmp2_" "// $e" 0
			return 0
		}
		if [catch {set sim_fd [open _tmp2_ r]} e] {
			show_log "error: $e\n"
			return 0
	}	}

	if {[gets $sim_fd line] > -1} {
		if {[string first "trail ends" $line] > 0} {
			executes "none" "-end-"
			catch { close $sim_fd }
			set sim_fd -1
			show_file "_tmp2_" "trail ends" 1
			.b.notes.t yview -pickplace end
			.b.notes.t insert end "\n(Use Up and Down arrows to move through trail)"
			.b.notes.t yview -pickplace end
			return 0
		}
		set f [split $line ","]
		if {[llength $f] >= 3} {
			set y [lindex $f 1]	;# '#' or '@'
			switch -- $y {
			"#"	{ set p [lindex $f 0]
				  set c [lindex $f 2]
				  executes $p $c
				}
			"@"	{ set p [lindex $f 0]
				  set c [lindex $f 2]
				  mark_state $p $c 1
		}	}	}
		return 1
	} else {
		executes "none" "-end-"
		.b.notes.t yview -pickplace end
		.b.notes.t insert end "\n(Use Up and Down arrows to move through trail)"
		catch { close $sim_fd }
		set sim_fd -1
		return 0
	}
}

proc do_sim {} {

	reset_states
	simulation_control
}

proc has_phrase {w s} {
	return [string first "$s" [$w get 0.0 end]]
}

proc do_check {} {
	global cnt state_exists statename declarations fname

	set fname [do_pml 1]
	if {$fname == ""} { return }

	set args "-run -bfs"	;# unless overridden below

	if {[string first "ltl" $declarations] >= 0} {
		# to get _spin_nvr.tmp
		show_log "spin -a $fname\n"
		catch "eval exec \"spin -a $fname > _tmp2_\"" e
		if {$e != ""} {
			set f "// $e"
			show_file "_tmp2_" "$f" 0
			catch { eval exec "rm -f pan.?" }
			return
		}
		if [catch {set fd [open "_spin_nvr.tmp" r]} e] {
			show_log "$e\n"
			return
		}
		while {[gets $fd line] > -1} {
			if {[string first "accept" $line] >= 0 \
			&&  [string first "accept_all" $line] < 0} {
				set args "-run -a"
				break
		}	}
		close $fd
	}

	if {$args == "-run -bfs"} {	;# check the statenames
		for {set i 0} {$i <= $cnt} {incr i} {
			if {[info exists state_exists(S$i)]} {
				# puts "$i	$statename($i)"
				if {[string first "accept" $statename($i)] == 0} {
					set args "-run -a"
					break
				}
				if {[string first "progress" $statename($i)] == 0} {
					set args "-run -DNP -l"
					break
	}	}	}	}

	show_log "rm -f ${fname}.trail\n"
	catch " eval exec rm -f ${fname}.trail"

	set f ""
	show_log "spin $args $fname\n"
	update

	catch " eval exec \"spin $args $fname > _tmp2_\" " e
	if {$e != ""} { show_log "$e\n"; set f "// $e" }
	show_file "_tmp2_" "$f" 0
	catch { eval exec "rm -f _tmp2_ pan" }
}

proc sanity_check {} {

	if {[auto_execok spin] == "" \
	||  [auto_execok spin] == 0} {
		puts "error: no executable Spin found..."
		puts "tau requires Spin Version 6 or later"
		exit 0
	} else {
		catch { set fd [open "|spin -V" r] } errmsg
		if {$fd == -1} {
			puts "$errmsg"
			exit 0
		} else {
			set version "Spin Version unknown"
			if {[gets $fd line] > -1} {
				set version "$line"
			}
			catch "close $fd"
		}
		if {[string first "Spin Version "  $version] < 0 \
		||  [string first "Spin Version 5" $version] >= 0 \
		||  [string first "Spin Version 4" $version] >= 0 \
		||  [string first "Spin Version 3" $version] >= 0 } {
			puts "error: tau requires Spin Version 6 or later"
			puts "you have: $version"
			exit 0
	}	}

	if {[auto_execok gcc] == "" \
	||  [auto_execok gcc] == 0} {
		puts "error: no C compiler (gcc) found..."
		puts "tau, and background tool Spin, needs a C compiler to work"
		exit 0
	}
	show_log "$version"
}

sanity_check

if {$argc == 1} {
	if {[string first ".tau" $argv] < 0} {
		puts "error: filename must have extension .tau"
		exit 0
	}
	openfile "$argv"
	set Fname "$argv"
}

update
